# Copyright 2020-2021 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Client log handler module."""
import os
import logging
import json
import shutil
from requests.exceptions import HTTPError
from etos_client.lib.graphql import request_artifacts

_LOGGER = logging.getLogger(__name__)


class ETOSLogHandler:
    """ETOS client log handler. Download all logs sent via EiffelTestSuiteFinishedEvent."""

    def __init__(self, etos, events):
        """Initialize log handler.

        :param etos: ETOS Library instance.
        :type etos: :obj:`etos_lib.etos.ETOS`
        :param events: All events collected from the test execution.
        :type events: list
        """
        self.etos = etos
        self.events = events

        self.report_dir = os.path.join(
            self.etos.config.get("workspace"), self.etos.config.get("report_dir")
        )
        self.artifact_dir = os.path.join(
            self.etos.config.get("workspace"), self.etos.config.get("artifact_dir")
        )
        self.prepare()

    def prepare(self):
        """Prepare the workspace for logs."""
        if not os.path.exists(self.report_dir):
            os.makedirs(self.report_dir)
        if not os.path.exists(self.artifact_dir):
            os.makedirs(self.artifact_dir)

    @staticmethod
    def _logs(test_suite_finished):
        """Iterate over all persistentLogs in test_suite_finished event.

        :param test_suite_finished: JSON data from test_suite_finished event.
        :type test_suite_finished: str
        :return: Log name and log url.
        :rtype: tuple
        """
        for log in test_suite_finished.get("data", {}).get(
            "testSuitePersistentLogs", []
        ):
            yield log.get("name"), log.get("uri")

    @property
    def all_logs(self):
        """Iterate over all logs for the executed test suite."""
        for finished in self.events.get("testSuiteFinished", []) + self.events.get(
            "mainSuiteFinished", []
        ):
            for log in self._logs(finished):
                yield log

    @property
    def all_artifacts(self):
        """Iterate over all artifacts for the executed test suite."""
        for artifact_created in request_artifacts(
            self.etos, self.events.get("activityId")
        ):
            for _, location in self.etos.utils.search(artifact_created, "uri"):
                suite_name = ""
                for link in artifact_created.get("links", []):
                    for _, name in self.etos.utils.search(
                        link.get("links", {}), "name"
                    ):
                        suite_name = name  # There should be exactly one!
                for _, name in self.etos.utils.search(
                    artifact_created.get("data", {}), "name"
                ):
                    yield f"{suite_name}_{name}", f"{location}/{name}"

    def _iut_data(self, environment):
        """Get IUT data from Environment URI.

        :param environment: Environment event to get URI from.
        :type environment: dict
        :return: IUT JSON data.
        :rtype: dict
        """
        if environment.get("uri"):
            iut_data = self.etos.http.wait_for_request(environment.get("uri"))
            return iut_data.json()
        return None

    @property
    def iuts(self):
        """All IUT Data environment events."""
        for environment in self.events.get("environmentDefined", []):
            if environment.get("data", {}).get("name", "").startswith("IUT Data"):
                yield self._iut_data(environment.get("data"))

    def _download(self, name, uri, directory, spinner):
        """Download a file and and write to disk.

        :param name: Name of resulting file.
        :type name: str
        :param uri: URI from where the file can be downloaded.
        :type uri: str
        :param directory: Into which directory to write the downloaded file.
        :type directory: str
        :param spinner: Spinner text item.
        :type spinner: :obj:`Spinner`
        """
        index = 0
        download_name = name
        while os.path.exists(os.path.join(directory, download_name)):
            index += 1
            download_name = f"{index}_{name}"
        spinner.text = f"Downloading {download_name}"
        generator = self.etos.http.wait_for_request(uri, as_json=False, stream=True)
        try:
            for response in generator:
                with open(os.path.join(directory, download_name), "wb+") as report:
                    for chunk in response:
                        report.write(chunk)
                break
            return True
        except (ConnectionError, HTTPError) as error:
            spinner.warn(f"Failed in downloading {download_name!r}.")
            spinner.warn(str(error))
            return False

    def download_logs(self, spinner):
        """Download all logs to report and artifact directories."""
        nbr_of_logs_downloaded = 0
        incomplete = False

        for name, uri in self.all_logs:
            result = self._download(name, uri, self.report_dir, spinner)
            if result:
                nbr_of_logs_downloaded += 1
            else:
                incomplete = True

        for name, uri in self.all_artifacts:
            result = self._download(name, uri, self.artifact_dir, spinner)
            if result:
                nbr_of_logs_downloaded += 1
            else:
                incomplete = True

        for index, iut in enumerate(self.iuts):
            if iut is None:
                break
            spinner.text = "Downloading IUT Data"
            try:
                filename = f"IUT_{index}.json"
                with open(
                    os.path.join(self.artifact_dir, filename), "w+", encoding="utf-8"
                ) as report:
                    json.dump(iut, report)
            except Exception as error:  # pylint:disable=broad-except
                spinner.warn(f"Failed in downloading {filename!r}.")
                spinner.warn(str(error))
                incomplete = True
            nbr_of_logs_downloaded += 1

        shutil.make_archive(
            os.path.join(self.artifact_dir, "reports"), "zip", self.report_dir
        )
        spinner.info(f"Downloaded {nbr_of_logs_downloaded} logs")
        spinner.info(f"Reports: {self.report_dir}")
        spinner.info(f"Artifacs: {self.artifact_dir}")
        if incomplete:
            spinner.fail("Logs failed downloading.")
            return False
        spinner.succeed("Logs downloaded.")
        return True



# Copyright 2020-2021 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Client test result handler."""
import time
import logging
from .graphql import (
    request_activity,
    request_activity_canceled,
    request_test_suite_finished,
    request_test_suite_started,
    request_suite,
    request_announcements,
    request_environment,
)

_LOGGER = logging.getLogger(__name__)


class ETOSTestResultHandler:
    """Handle ETOS test results."""

    activity_id = None

    def __init__(self, etos):
        """Initialize ETOS Client test result handler.

        :param etos: ETOS Library instance.
        :type etos: :obj:`etos_lib.etos.ETOS`
        """
        self.etos = etos
        self.events = {}
        self.announcements = []

    @property
    def has_started(self):
        """Whether or not test suites have started.

        :return: Whether or not test suites have started.
        :rtype bool
        """
        return len(self.events.get("testSuiteStarted", [])) >= 1

    @property
    def has_finished(self):
        """Whether or not test suites have finished.

        :return: Whether or not test suites have finished.
        :rtype bool
        """
        return bool(self.events.get("mainSuiteFinished", False))

    @property
    def spinner_text(self):
        """Generate the spinner text based on test results.

        :return: String formatted for the Halo spinner.
        :rtype: str
        """
        message_template = (
            "{announcement}\t"
            "Started : {started_length}\t"
            "Finished: {finished_length}\t"
        )
        try:
            announcement = self.announcements[-1]["data"]["body"]
        except (KeyError, IndexError, TypeError):
            announcement = ""

        params = {
            "started_length": len(self.events.get("testSuiteStarted", [])),
            "finished_length": len(self.events.get("testSuiteFinished", [])),
            "announcement": announcement,
        }
        return message_template.format(**params)

    def get_environment_events(self):
        """Get the environment events set for this execution."""
        events = self.events.get("testSuiteStarted", [])
        ids = [
            started["meta"]["id"]
            for started in events + [self.events.get("mainSuiteStarted")]
        ]
        ids.append(self.activity_id)
        self.events["environmentDefined"] = list(request_environment(self.etos, ids))

    def test_result(self):
        """Build test results based on events retrieved.

        :return: Result and message.
        :rtype: tuple
        """
        nbr_of_fail = 0
        if not self.events.get("testSuiteStarted"):
            return False, "Test suite did not start."
        self.get_environment_events()

        main_suite_finished = self.events.get("mainSuiteFinished", [{}])[0]
        data = main_suite_finished.get("data", {})
        outcome = data.get("testSuiteOutcome", {})
        verdict = outcome.get("verdict")
        if verdict == "PASSED":
            result = True
        else:
            for test_suite_finished in self.events.get("testSuiteFinished"):
                data = test_suite_finished.get("data", {})
                outcome = data.get("testSuiteOutcome", {})
                verdict = outcome.get("verdict")
                if verdict != "PASSED":
                    nbr_of_fail += 1
        if nbr_of_fail > 0:
            nbr_of_suites = len(self.events.get("testSuiteStarted"))
            message = f"{nbr_of_fail}/{nbr_of_suites} test suites failed."
            result = False
        else:
            message = "Test suite finished successfully."
            result = True
        return result, message

    def latest_announcement(self, spinner):
        """Find latest announcement and print it.

        :param spinner: Spinner text item.
        :type spinner: :obj:`Spinner`
        """
        for announcement in request_announcements(
            self.etos, [self.etos.config.get("suite_id"), self.activity_id]
        ):
            if announcement not in self.announcements:
                self.announcements.append(announcement)
                data = self.announcements[-1].get("data")
                spinner.info(f"{data.get('heading')}: {data.get('body')}")
                spinner.start("Waiting for ETOS.")

    @staticmethod
    def split_main_and_sub_suites_started(started):
        """Split test suite started into main suite and sub suites.

        :param started: Test suites that have started in a specified activity.
        :type started: list
        :return: Main suite and sub suites started.
        :rtype: tuple
        """
        sub_suite_started = []
        main_suite = None
        for test_suite_started in started:
            for category in test_suite_started["data"]["testSuiteCategories"]:
                if "Sub suite" in category.get("type"):
                    sub_suite_started.append(test_suite_started)
                    break
            else:
                main_suite = test_suite_started
        return main_suite, sub_suite_started

    def get_events(self, suite_id):
        """Get events from an activity started from suite id.

        :param suite_id: ID of test execution recipe that triggered this activity.
        :type suite_id: str
        :return: Dictionary of all events generated for this suite.
        :rtype: dict
        """
        if self.activity_id is None:
            activity = request_activity(self.etos, suite_id)
            if activity is None:
                return {}
            self.activity_id = activity["meta"]["id"]

        results = {"activityId": self.activity_id}

        started = list(request_test_suite_started(self.etos, self.activity_id))
        if not started:
            return {}
        main_suite, sub_suite_started = self.split_main_and_sub_suites_started(started)
        results["testSuiteStarted"] = sub_suite_started
        results["mainSuiteStarted"] = main_suite

        started_ids = [
            test_suite_started["meta"]["id"] for test_suite_started in started
        ]
        main_suite_link = {
            "testSuiteStarted": {"meta": {"id": main_suite["meta"]["id"]}}
        }

        finished = list(request_test_suite_finished(self.etos, started_ids))
        if not finished:
            return results
        results["testSuiteFinished"] = [
            sub_finished
            for sub_finished in finished
            if main_suite_link not in sub_finished["links"]
        ]
        results["mainSuiteFinished"] = [
            main_suite
            for main_suite in finished
            if main_suite_link in main_suite["links"]
        ]
        return results

    def print_suite(self, spinner):
        """Print test suite batchesUri.

        :param spinner: Spinner text item.
        :type spinner: :obj:`Spinner`
        """
        spinner.text = "Waiting for test suite url."
        timeout = time.time() + 60
        while time.time() < timeout:
            tercc = request_suite(self.etos, self.etos.config.get("suite_id"))
            if tercc:
                spinner.info(f"Test suite: {tercc['data']['batchesUri']}")
                spinner.start()
                return
            time.sleep(1)
        raise TimeoutError("Test suite not available in 10s.")

    def wait_for_test_suite_finished(self, spinner):
        """Query graphql server until the number of started is equal to number of finished.

        :param spinner: Spinner text item.
        :type spinner: :obj:`Spinner`
        :return: Whether it was a successful execution or not and the test results.
        :rtype: tuple
        """
        self.print_suite(spinner)
        timeout = time.time() + 3600 * 24
        while time.time() < timeout:
            self.latest_announcement(spinner)
            time.sleep(10)
            self.events = self.get_events(self.etos.config.get("suite_id"))
            canceled = request_activity_canceled(self.etos, self.activity_id)
            if canceled:
                return False, canceled["data"]["reason"]
            if not self.has_started:
                continue
            spinner.text = self.spinner_text
            if self.has_finished:
                return self.test_result()
        return False, "Test suites did not finish"



# Copyright 2020-2021 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GraphQL query handler."""
from .graphql_queries import (
    ARTIFACTS,
    ACTIVITY_TRIGGERED,
    ACTIVITY_CANCELED,
    TEST_SUITE_STARTED,
    TEST_SUITE_FINISHED,
    TEST_SUITE,
    ANNOUNCEMENTS,
    ENVIRONMENTS,
)


def request(etos, query):
    """Request graphql in a generator.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param query: Query to send to graphql.
    :type query: str
    :return: Generator
    :rtype: generator
    """
    wait_generator = etos.utils.wait(etos.graphql.execute, query=query, timeout=60)
    yield from wait_generator


def request_suite(etos, suite_id):
    """Request a tercc from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param suite_id: ID of execution recipe.
    :type suite_id: str
    :return: Response from graphql or None
    :rtype: dict or None
    """
    for response in request(etos, TEST_SUITE % suite_id):
        try:
            _, tercc = next(
                etos.graphql.search_for_nodes(
                    response, "testExecutionRecipeCollectionCreated"
                )
            )
        except StopIteration:
            return None
        return tercc
    return None


def request_activity(etos, suite_id):
    """Request an activity event from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param suite_id: ID of execution recipe triggering this activity.
    :type suite_id: str
    :return: Response from graphql or None
    :rtype: dict or None
    """
    for response in request(etos, ACTIVITY_TRIGGERED % suite_id):
        if response:
            try:
                _, activity = next(
                    etos.graphql.search_for_nodes(response, "activityTriggered")
                )
            except StopIteration:
                return None
            return activity
    return None


def request_activity_canceled(etos, activity_id):
    """Request an activity event from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param activity_id: ID of the activity beeing canceled.
    :type activity_id: str
    :return: Response from graphql or None
    :rtype: dict or None
    """
    for response in request(etos, ACTIVITY_CANCELED % activity_id):
        if response:
            try:
                _, activity_canceled = next(
                    etos.graphql.search_for_nodes(response, "activityCanceled")
                )
            except StopIteration:
                return None
            return activity_canceled
    return None


def request_test_suite_started(etos, activity_id):
    """Request test suite started from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param activity_id: ID of activity in which the test suites started
    :type activity_id: str
    :return: Iterator of test suite started graphql responses.
    :rtype: iterator
    """
    for response in request(etos, TEST_SUITE_STARTED % activity_id):
        if response:
            for _, test_suite_started in etos.graphql.search_for_nodes(
                response, "testSuiteStarted"
            ):
                yield test_suite_started
            return None  # StopIteration
    return None  # StopIteration


def request_test_suite_finished(etos, test_suite_ids):
    """Request test suite finished from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param test_suite_ids: list of test suite started IDs of which finished to search for.
    :type test_suite_ids: list
    :return: Iterator of test suite finished graphql responses.
    :rtype: iterator
    """
    or_query = "{'$or': ["
    or_query += ", ".join(
        [
            f"{{'links.type': 'TEST_SUITE_EXECUTION', 'links.target': '{test_suite_id}'}}"
            for test_suite_id in test_suite_ids
        ]
    )
    or_query += "]}"
    for response in request(etos, TEST_SUITE_FINISHED % or_query):
        if response:
            for _, test_suite_finished in etos.graphql.search_for_nodes(
                response, "testSuiteFinished"
            ):
                yield test_suite_finished
            return None  # StopIteration
    return None  # StopIteration


def request_announcements(etos, ids):
    """Request announcements from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param ids: list of IDs of which announcements to search for.
    :type ids: list
    :return: Iterator of announcement published graphql responses.
    :rtype: iterator
    """
    or_query = "{'$or': ["
    or_query += ", ".join([f"{{'links.target': '{_id}'}}" for _id in ids])
    or_query += "]}"
    for response in request(etos, ANNOUNCEMENTS % or_query):
        if response:
            for _, announcement in etos.graphql.search_for_nodes(
                response, "announcementPublished"
            ):
                yield announcement
            return None  # StopIteration
    return None  # StopIteration


def request_environment(etos, ids):
    """Request environments from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param ids: list of IDs of which environments to search for.
    :type ids: list
    :return: Iterator of environnent defined graphql responses.
    :rtype: iterator
    """
    or_query = "{'$or': ["
    or_query += ", ".join([f"{{'links.target': '{_id}'}}" for _id in ids])
    or_query += "]}"
    for response in request(etos, ENVIRONMENTS % or_query):
        if response:
            for _, environment in etos.graphql.search_for_nodes(
                response, "environmentDefined"
            ):
                yield environment
            return None  # StopIteration
    return None  # StopIteration


def request_artifacts(etos, context):
    """Request artifacts from graphql.

    :param etos: Etos Library instance for communicating with ETOS.
    :type etos: :obj:`etos_lib.etos.ETOS`
    :param context: ID of the activity used in CONTEXT.
    :type context: str
    """
    for response in request(etos, ARTIFACTS % context):
        if response:
            for _, artifact in etos.graphql.search_for_nodes(
                response, "artifactCreated"
            ):
                yield artifact
            return None  # StopIteration
    return None  # StopIteration



# -*- coding: utf-8 -*-
# Copyright 2020-2021 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Main executable module."""

import argparse
import sys
import os
import logging
import time
import json

import requests
from halo import Halo
from etos_lib.etos import ETOS

from etos_client import __version__
from etos_client.client import ETOSClient
from etos_client.lib import ETOSTestResultHandler, ETOSLogHandler

_logger = logging.getLogger(__name__)


def environ_or_required(key):
    """Get key from environment or make it a required input.

    If a key is available as an environment variable, use it as default.
    Else require the argument to be set.

    Note:
        It is still possible to override the value with input even if the
        environment variable is set.
    """
    if os.getenv(key):
        return {"default": os.getenv(key)}
    return {"required": True}


def parse_args(args):
    """Parse command line parameters.

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Client for executing test automation suites in ETOS"
    )
    parser.add_argument(
        "cluster",
        help="Cluster is the URL to the ETOS API.",
    )
    parser.add_argument(
        "-i",
        "--identity",
        help="Artifact created identity purl or ID to execute test suite on.",
        **environ_or_required("IDENTITY"),
    )
    parser.add_argument(
        "-s",
        "--test-suite",
        help="Test suite execute. Either URL or name.",
        **environ_or_required("TEST_SUITE"),
    )

    parser.add_argument(
        "--no-tty", help="Disable features requiring a tty.", action="store_true"
    )
    parser.add_argument(
        "-w",
        "--workspace",
        default=os.getenv("WORKSPACE", os.getcwd()),
        help="Which workspace to do all the work in.",
    )
    parser.add_argument(
        "-a",
        "--artifact-dir",
        default="artifacts",
        help="Where test artifacts should be stored. Relative to workspace.",
    )
    parser.add_argument(
        "-r",
        "--report-dir",
        default="reports",
        help="Where test reports should be stored. Relative to workspace.",
    )
    parser.add_argument(
        "-d",
        "--download-reports",
        default=None,
        help="Should we download reports. Can be 'yes', 'y', 'no', 'n'.",
    )

    parser.add_argument(
        "--iut-provider", default="default", help="Which IUT provider to use."
    )
    parser.add_argument(
        "--execution-space-provider",
        default="default",
        help="Which execution space provider to use.",
    )
    parser.add_argument(
        "--log-area-provider", default="default", help="Which log area provider to use."
    )
    parser.add_argument(
        "--dataset",
        default="{}",
        help=(
            "Additional dataset information to the environment provider. "
            "Check with your provider which information can be supplied."
        ),
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"etos_client {__version__}",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        default=logging.WARNING,
        const=logging.INFO,
    )
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        default=logging.WARNING,
        const=logging.DEBUG,
    )
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Set up basic logging.

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


def check_etos_connectivity(url):
    """Check the connection to ETOS.

    :param url: URL to test connection to.
    :type url: str
    """
    try:
        response = requests.get(url, timeout=5)
        response.raise_for_status()
    except Exception as exception:  # pylint:disable=broad-except
        raise Exception(
            "Unable to connect to ETOS. Please check your network connection."
        ) from exception


class Printer:
    """Generic printing class matching the interface of halo."""

    __text = None
    timeout = 60
    timer = None

    def __init__(self, text, **_):
        """Initialize the printer class.

        :param text: Initial text.
        :type text: str
        """
        self.timer = time.time() + self.timeout
        self.text = text

    def __enter__(self):
        """Enter printer context."""
        return self

    @property
    def text(self):
        """Printer text."""
        return self.__text

    @text.setter
    def text(self, text):
        """Set text for this object.

        Don't reprint unless the text is new or a timeout has been reached.
        This is to disable unecessary text spam into the console log.

        :param text: Text to print.
        :type text: str
        """
        timeout_reached = time.time() > self.timer
        if text != self.__text or timeout_reached:
            self.timer = time.time() + self.timeout
            self.info(text)
        self.__text = text

    @staticmethod
    def info(text):
        """Print text to stdout.

        :param text: Text to print.
        :type text: str
        """
        _logger.info(text)

    @staticmethod
    def fail(text):
        """Print text to stdout.

        :param text: Text to print.
        :type text: str
        """
        _logger.error(text)

    @staticmethod
    def warn(text):
        """Print text to stdout.

        :param text: Text to print.
        :type text: str
        """
        _logger.warning(text)

    succeed = info

    def start(self, text=None):
        """Set starting text."""
        if text is not None:
            self.text = text

    def __exit__(self, *_, **__):
        """Exit printer context."""


def generate_spinner(no_tty):
    """If no tty, return a generic 'printer' class. Else return halo.

    :param no_tty: Whether the executor has an interactive tty or not.
    :type no_tty: bool
    :return: Spinner text item.
    :rtype: :obj:`Spinner`
    """
    if no_tty:
        return Printer
    return Halo


def main(args):
    """Entry point allowing external calls.

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    etos = ETOS("ETOS Client", os.getenv("HOSTNAME"), "ETOS Client")

    setup_logging(args.loglevel)
    info = generate_spinner(args.no_tty)

    for key, value in args._get_kwargs():  # pylint:disable=protected-access
        etos.config.set(key, value)

    etos.config.set("artifact_identifier", args.identity)
    etos.config.set("dataset", json.loads(args.dataset))

    with info(text="Checking connectivity to ETOS", spinner="dots") as spinner:
        spinner.info(f"Running in cluster: {args.cluster!r}")
        spinner.info("Configuration:")
        spinner.info(f"{etos.config.config}")
        try:
            check_etos_connectivity(f"{args.cluster}/selftest/ping")
        except Exception as exception:  # pylint:disable=broad-except
            spinner.fail(str(exception))
            sys.exit(1)
        spinner.start()
        spinner.succeed("Connection successful.")
        spinner.succeed("Ready to launch ETOS.")

        # Start execution
        etos_client = ETOSClient(etos, args.cluster)
        spinner.start("Triggering ETOS.")
        success = etos_client.start(spinner)
        if not success:
            # Unix  : 0 == Success, 1 == Fail
            # Python: 1 == True   , 0 == False
            sys.exit(not success)

        spinner.info(f"Suite ID: {etos_client.test_suite_id}")
        spinner.info(f"Artifact ID: {etos_client.artifact_id}")
        spinner.info(f"Purl: {etos_client.artifact_identity}")

        etos.config.set("suite_id", etos_client.test_suite_id)
        os.environ["ETOS_GRAPHQL_SERVER"] = etos_client.event_repository
        spinner.info(f"Event repository: {etos.debug.graphql_server!r}")

        # Wait for test results
        test_result_handler = ETOSTestResultHandler(etos)
        spinner.start("Waiting for ETOS.")
        success, results = test_result_handler.wait_for_test_suite_finished(spinner)
        if not success:
            spinner.fail(results)
        else:
            spinner.succeed(results)

        # Download reports
        if args.download_reports is None:
            answer = input(
                "Do you want to download all logs for this test execution? (y/n): "
            )
            while answer.lower() not in ("y", "yes", "no", "n"):
                print("Please answer 'yes' or 'no'")
                answer = input()
        else:
            answer = args.download_reports
        if answer.lower() in ("y", "yes"):
            log_handler = ETOSLogHandler(etos, test_result_handler.events)
            spinner.start("Downloading test logs.")
            logs_downloaded_successfully = log_handler.download_logs(spinner)
            if not logs_downloaded_successfully:
                sys.exit("ETOS logs did not download successfully.")


def run():
    """Entry point for console_scripts."""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# Copyright 2020-2021 Axis Communications AB.
#
# For a full list of individual contributors, please see the commit history.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ETOS Client module."""
import logging
from uuid import UUID
from packageurl import PackageURL

_LOGGER = logging.getLogger(__name__)


class ETOSClient:
    """Client for starting test suites in ETOS."""

    event_repository = None
    test_suite_id = None
    artifact_id = None
    artifact_identity = None

    def __init__(self, etos, cluster):
        """Initialize ETOS client.

        :param etos: ETOS Library instance.
        :type etos: :obj:`etos_lib.etos.ETOS`
        :param cluster: ETOS cluster to start tests in.
        :type cluster: str
        """
        self.etos = etos
        self.cluster = cluster
        self.test_execution = {}

    @staticmethod
    def is_uuid(string):
        """Test if string is a valid UUID v4.

        :param string: A string to test.
        :type string: str
        :return: Is string a valid UUID.
        :rtype: bool
        """
        try:
            UUID(string, version=4)
            return True
        except ValueError:
            return False

    @staticmethod
    def is_packageurl(string):
        """Test if string is a valid PackageURL.

        :param string: A string to test.
        :type string: str
        :return: Is string a valid PackageURL.
        :rtype: bool
        """
        try:
            PackageURL.from_string(string)
            return True
        except ValueError:
            return False

    @property
    def data(self):
        """ETOS request data."""
        data = {
            "dataset": self.etos.config.get("dataset"),
            "iut_provider": self.etos.config.get("iut_provider"),
            "execution_space_provider": self.etos.config.get(
                "execution_space_provider"
            ),
            "log_area_provider": self.etos.config.get("log_area_provider"),
            "test_suite_url": self.etos.config.get("test_suite"),
        }
        artifact_identifier = self.etos.config.get("artifact_identifier")
        if self.is_uuid(artifact_identifier):
            data["artifact_id"] = artifact_identifier
        elif self.is_packageurl(artifact_identifier):
            data["artifact_identity"] = artifact_identifier
        else:
            raise ValueError(
                f"Identity {artifact_identifier!r} is not a valid PackageURL or UUID."
            )
        return data

    def start(self, spinner):
        """Start ETOS test execution.

        :param spinner: Spinner text item.
        :type spinner: :obj:`Spinner`
        :return: Whether or not suite triggered correctly.
        :rtype: bool
        """
        spinner.info(str(self.data))
        generator = self.etos.http.retry(
            "POST", f"{self.cluster}/etos", timeout=30, json=self.data
        )
        response = None
        try:
            for response in generator:
                self.test_execution = response
                self.test_suite_id = response.get("tercc")
                self.artifact_id = response.get("artifact_id")
                self.artifact_identity = response.get("artifact_identity")
                self.event_repository = response.get("event_repository")
                break
        except ConnectionError as exception:
            spinner.warn(str(exception))
            spinner.fail("Failed to trigger ETOS.")
            return False
        spinner.succeed("ETOS triggered.")
        return True
